/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "fitsio.h"
#include "imaging_fdata.h"
#define clight  2.9979250e8 /*m per sec */
#define twopi   6.2831853072
#ifdef HP
#define finite(x) isfinite(x)
#endif

/*
    Usage::
    oirRotateInsOpd infile outfile  [coordRefFile]
*/
void Usage() {
   printf("   oirRotateInsOpd inFile outFile [coordRefFile]\n");
   printf("   inFile = file name of output of oirFormFringes\n");
   printf("   outFile = output file name\n");
   printf("   coordRefFile = name of FITS file from which to extract wavelength information\n");
   printf("       if the inFile information is inaccurate.\n");
}

/****************************************************************
  *  find position and value of minimum and maximum in an array
  *
  *  INPUTS:
  *    a   float *  array of input values
  *    n   int      number of elements in array
  *
  *  OUTPUTS:
  *    min float *  address to receive minimum value in array
  *    max float *  address to receive maximum value in array
  *    imin int  *  address to receive index of minumum value
  *    imax int  *  address to receive index of maximum value
  ***************************************************************/
void mmax(float *a, int n, float *min, float *max, int *imin, int *imax) {
   int i,j,k;
   float in,ax;
   j=0;
   k=0;
   in = a[0];
   ax = in;
   for (i=0;i<n;i++){
      if (a[i] > ax) {
         ax = a[i];
         j  = i;
      } 
      if (a[i] < in) {
         in = a[i];
         k  = i;
      } 
   }
   *imin = k;
   *imax = j;
   *min  = in;
   *max  = ax;
   return;
}
/****************************************************************
 *      determine the arithmetic mode of a group of data
 *      by binning into a histogram and finding the position
 *      of max of histogram
 *
 *   INPUTS:
 *     nData   int     number of data points
 *     nBin    int     number of bins to use in histogram
 *     data    float*  input data
 *
 *   RETURNS:
 *             float   value of bin that contains the most data points
 *********************************************************************/
float oirMode(int nData, int nBin, float *data) {

   float *hist;
   int iMin, iMax;
   float dMin, dMax, dData;
   float hMin, hMax ;
   float mode;
   int i,j;

/*            
              find min and max of data values    
              allocate histogram array and clear
*/
   mmax(data, nData, &dMin, &dMax, &iMin, &iMax);
   dData = (dMax-dMin)/nBin;
   hist = (float *)malloc((nBin+1) * sizeof(int));
   memset(hist, 0, (nBin+1)*sizeof(int));
/*
              put data points into histogram
*/
   for (i=0; i<nData-1; i++) {
      j = (data[i]-dMin)/dData;
      hist[j] = hist[j]+1.;
   }
/*
              find max of histogram and clean up
*/
   mmax(hist, nBin+1, &hMin, &hMax, &iMin, &iMax);
   mode = dMin + iMax * dData;

   free(hist);
   return mode;
}  /* end of oirMode */

/*************************************************************************
 *    Remove phase from known instrumental opd components to MIDI data
 *    i.e. find the net recorded [opd + local_opd] in the data stream
 *    and multiply each point by exp(-i * opd * k) where k is 2pi/lambda.
 *    Be a little careful: the absolute zero point of the opd system is
 *    undefined and this could result in very large values (e.g. meters)
 *    so that very small errors in lambda would cause large phase errors.
 *    So assume that the on-line fringe tracking did something reasonable,
 *    and use a provisional opd zero point of the arithmetic mode of the
 *    data points.  In any case write this value in the header of the
 *    output data table as OPD0 so we know what we did.
 *
 *    The output data is complex; we use a pseudo-complex (i.e. pairs
 *    [Real, Imaginary]) instead of true FITS complex values.
 *
 *    Finally, allow the possibility that someone has a better wavelength
 *    calibration that that stored with the original data, so optionally
 *    read this wavelength data in from a coordinate reference file.
*************************************************************************/
int main (int argc, char** argv) {

   int i, j, ierr;
   long nRow, iRow, jRow ;
   int nx ;
   float *frequency, *waveNo;
   double opd0, opd;
   float *opdList;
   float opdMode;

   char outFileName[130];
   char *bang="!";
   char history[80];
   char nrtsMode[80];
   char *offsetTracking;

   fitsfile *inFile;
   fitsfile *refFile;
   fitsfile *outFile = NULL;
   int iKey, maxstep;
   imaging_fdata_table_def *idata;
   imaging_fdata_table_def *odata;
   imaging_detector_table_def *outDetectorDef;
   imaging_detector_table_def *refDetectorDef;
   imaging_fdata_row *inData;
   imaging_fdata_row *outData;
   int maxCo, odim;

   ierr = 0;

   if (argc < 2) {
      Usage();
      return 1;
   }
   printf("      ****** oirRotateInsOpd ******\n");
/*
      open input file and get detector info
*/
   fits_open_file (&inFile, argv[1], READONLY, &ierr);
   if (ierr != 0) {
      printf("ERROR:fits openfile of %s gave error %i\n",argv[1],
         ierr);
      printf("%s \n",strerror(errno));
      return 1;
   }
   i = fits_read_key(inFile, TSTRING, "HIERARCH ESO DET NRTS MODE", nrtsMode,0,&ierr);
   offsetTracking = strstr(nrtsMode, "OFF");
/*
      create output data table 
*/
   strcpy(outFileName, bang);
   strcat(outFileName, argv[2]);

   fits_create_file (&outFile, outFileName, &ierr);
   if (ierr !=0) {
      printf ("ERROR: opening fits output %s %i\n",argv[2], ierr);
       printf("%s \n",strerror(errno));
      return 1;
      }
/*   copy header info */
   i = fits_copy_header(inFile, outFile, &ierr);
   snprintf(history,80,"oirRotateInsOpd INPUT %s", argv[1]);
   i=fits_write_history(outFile, history, &ierr);
   if (argc > 3) {
      snprintf(history,80,"oirRotateInsOpd COORD REFERENCE FILE %s", argv[3]);
   }

/*   get detector info from input file */
   outDetectorDef = read_imaging_detector_table(inFile, &ierr);
   idata          = open_imaging_fdata_table(inFile, &ierr);
   iKey = table_find_key(idata->def, "MAXSTEP");
   if (iKey>0)  table_get_keyword (idata->def, iKey, (void*)&maxstep,
      &ierr);
/*
      if reference coord system specified, copy coord data into output table
*/
   if (argc > 3) {
      fits_open_file (&refFile, argv[3], READONLY, &ierr);
      if (ierr != 0) {
         printf("ERROR:fits reference open file of %s gave error %i\n",argv[3],
            ierr);
         printf("%s \n",strerror(errno));
         return 1;
      }
      refDetectorDef = read_imaging_detector_table(refFile, &ierr);
      if (refDetectorDef->naxis[1][0] != outDetectorDef->naxis[0][0]) {
         printf("nx in input data %i != reference data %i \n", 
            outDetectorDef->naxis[1][0],refDetectorDef->naxis[1][0]);
         return 1;
      }

      fits_close_file(refFile, &ierr);
      maxCo = refDetectorDef->max_coef;
      if (maxCo > outDetectorDef->max_coef) {
         printf("maxCoeff in ref detector %i > data %i\n",
            maxCo, outDetectorDef->max_coef);
         return 1;
      }
      for (i=0;i<2;i++) {
         outDetectorDef->crpix[0][i] = refDetectorDef->crpix[1][i];
         outDetectorDef->crval[0][i] = refDetectorDef->crval[1][i];
      }
      for (odim = 0; odim < 2; odim++) for (i=0; i< maxCo; i++) {
         outDetectorDef->dmc[0][i][odim] = refDetectorDef->dmc[1][i][odim];
         outDetectorDef->dmp[0][i][0][odim] = refDetectorDef->dmp[1][i][0][odim];
         outDetectorDef->dmp[0][i][1][odim] = refDetectorDef->dmp[1][i][1][odim];
      }
      for (i=0;i<4;i++) outDetectorDef->cd[0][i] = refDetectorDef->cd[1][i];
   }
/*     
         get row size
*/
   nx = outDetectorDef->naxis[0][0];
/*       
         get frequency information from detector table
*/
   frequency = imaging_detector_wcs(outDetectorDef, 1);
   waveNo = (float *)malloc(nx * sizeof(float));
   for (i=0; i<nx; i++) waveNo[i] = twopi * frequency[i]/clight;
/*     
         double row size (pseudo-complex)
*/
   outDetectorDef->naxis[0][0] = 2*nx; 

   odata = make_imaging_fdata_table_def (outFile, 
      outDetectorDef, maxstep, &ierr); 
   if (ierr !=0) {
      printf ("ERROR: creating data definition\n");
      return 1;
      }
   inData = make_imaging_fdata_row(idata);
   outData = make_imaging_fdata_row(odata);
   i = fits_get_num_rows (inFile,  &nRow, &ierr); 
   opdList = (float *)malloc(nRow * sizeof(float));
   jRow = 0;
   for (iRow = 1; iRow <= nRow; iRow++) {
      i = read_imaging_fdata_table(inFile, idata, inData, iRow, &ierr);
      if (finite(inData->time))
         opdList[jRow++] = inData->opd[1]+inData->local_opd[1] - inData->opd[0] -
            inData->local_opd[0];
   }
/*
 *           compute mode but drop 1st 800 points because they were not tracking
*/
   opdMode = oirMode(jRow-800, 50, opdList+800);
   free(opdList);
   printf("   opd mode                     =%9.3f mm\n", opdMode*1.e3);
   opd0 = opdMode;
/*
 *       if offsetTracking was imposed, decrease opd0 by 50 microns
*/
   if (offsetTracking != NULL) opd0 = opd0 - 50.e-6; 
   printf("   opd0                         =%9.3f mm\n", opd0*1.e3);
   i = fits_write_key (outFile, TDOUBLE, "OPD0", (void*)(&opd0), "OFFSET TO OPD", &ierr);
/*       
         now loop over input data, multiplying input value * exp(i*k*(D-opd0))
*/
   for (iRow=1; iRow<=nRow; iRow++) {
/*
        read new row, multiply by cos/sin of phase and store
*/
      i = read_imaging_fdata_table(inFile, idata, inData, iRow, &ierr);

/*
            copy nonData info
*/
      outData->time = inData->time;
      outData->opd[0] = inData->opd[0];
      outData->opd[1] = inData->opd[1];
      outData->local_opd[1] = inData->local_opd[1];
      outData->local_opd[0] = inData->local_opd[0];
      outData->exp = inData->exp;
      outData->offset[0] = inData->offset[0];
      outData->offset[1] = inData->offset[1];
      outData->rotation  = inData->rotation;
      outData->frame     = inData->frame;
      outData->step      = inData->step ;
      outData->reference = inData->reference;
      outData->targtype[0] = inData->targtype[0];  
/*
         calculate opd, remove zero point, rotate fringe
*/
      opd = inData->opd[1]+inData->local_opd[1] - inData->opd[0] - inData->local_opd[0];
      opd = opd - opd0;
      for (j=0; j<nx; j++) {
         outData->data[0][2*j]   = inData->data[0][j] * cos(opd * waveNo[j]);
         outData->data[0][2*j+1] = inData->data[0][j] * sin(opd * waveNo[j]);
      }
      i = write_imaging_fdata_table(outFile, odata, outData, iRow, &ierr);
   }  /*  loop over input rows */
   i = close_imaging_fdata_table(outFile, odata, &ierr);
   i = close_imaging_fdata_table(inFile, idata, &ierr);
   kill_imaging_fdata_row(outData);
   kill_imaging_fdata_row(inData);
   free(frequency);
   free(waveNo);
   return 0;
}
